//======================================================================
//
//                    admdriver.c
//
// Purpose:
//          - verify, compare and time TAC++ generated adjoint code against
//            hand-written code by Mike Giles.
//
//======================================================================

/* Program to compute deltas and vegas of swaption portfolio
   from forward and reverse mode pathwise sensitivities 
   written by Mike Giles in 2005-7 based on code written by
   Zhao and Glasserman at Columbia University */


#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include "tac++-codes.h"
#include "mikes-codes.h"

#undef  PORTFOLIO_ADM_TIMING    // debugging only

/* -------------------------------------------------------- */

int main(){

    double delta = 0.25; /* LIBOR interval  */

    // data for swaption portfolio //
    int    Nopt = 15;
    int    maturities[] = {4,4,4,8,8,8,20,20,20,28,28,28,40,40,40};
    double swaprates[]  = {.045,.05,.055,.045,.05,.055,.045,.05,
			   .055,.045,.05,.055,.045,.05,.055 };

    int       i, N, Nmat, pass, path, npath;
    double    *v, *v_d, *L, *L_b, *L_d, *L2, *L3, 
	*lambda, *lambda_b, *z, tt[6];

    Nmat = 40;
    N = Nmat+40;

    v   = (double*) malloc(sizeof(double));
    L   = (double*) malloc(N*sizeof(double));
    L_b = (double*) malloc(N*sizeof(double));
    v_d = (double*) malloc(2*N*sizeof(double));
    L_d = (double*) malloc(2*N*N*sizeof(double));
    L2  = (double*) malloc(N*N*sizeof(double));
    L3  = (double*) malloc(N*N*sizeof(double));

    lambda   = (double*) malloc(N*sizeof(double));
    lambda_b = (double*) malloc(N*sizeof(double));
    z        = (double*) malloc(N*sizeof(double));

    //------------------------------
    // variables required for running
    // the tac(++) generated code
    //------------------------------
    double * L_ad;
    double * lambda_ad;
    double *vptr;
    double *vptr_ad;
    vptr      = (double*) malloc(sizeof(double));
    vptr_ad   = (double*) malloc(sizeof(double));
    L_ad      = (double*) malloc(N*sizeof(double));
    lambda_ad = (double*) malloc(N*sizeof(double));

    //============================================================
    //
    //          verify portfolio adjoint sensitivites on its own
    //
    //============================================================
    for(i=0;i<N;i++) L[i] = 0.05;
    portfolio_b(N,Nmat,delta,Nopt,maturities,swaprates,L,L_b,v);

    for(i=0;i<N;i++) L[i] = 0.05;
    for(i=0;i<N;i++) L_ad[i]=0;
    *vptr_ad = 1;
    tac_portfolio_ext_ad(N,Nmat,delta,Nopt,maturities,swaprates,L, L_ad, vptr, vptr_ad);

    printf("\n==================================================\n");
    printf("  portfolio sensitivity check, hand vs. tac(++)\n");
    printf("==================================================\n");
    printf( "%4sd %14s  %14s  %14s  \n",
	    "idx", "hand ADM", "TAC(++) ADM ", "rel. Diff");
    for(i=0; i<N; i++) {
	printf("%4d  %14.8f  %14.8f  %14.8f\n", i, L_b[i], L_ad[i], reldiff(L_b[i],L_ad[i], epsmach));
    }




    //============================================================
    //
    //          portfolio timing
    //
    //============================================================
#ifdef PORTFOLIO_ADM_TIMING
    for (pass=0; pass<6; pass++) {
	if (pass==0)
	    npath = 1;
	else if (pass<6)
	    npath = 100000;

	for (path=0; path<npath; path++) {

	    if (pass==1) {
		// plain evaluation for timing comparison //
		for(i=0;i<N;i++) {
		    L[i] = 0.05;
		}
		portfolio(N,Nmat,delta,Nopt,maturities,swaprates,L,v);
	    }
	    if (pass==0 | pass==2) {
		// hand-coded reverse mode //
		for(i=0;i<N;i++) {
		    L[i] = 0.05;
		}
		portfolio_b(N,Nmat,delta,Nopt,maturities,swaprates,L,L_b,v);
	    }

	    if (pass==0 | pass==3) {
		// hand-coded forward mode //
		for(i=0;i<N;i++) L[i] = 0.05;
		for(i=0;i<N*N;i++) L_d[i] = 0;
		for(i=0;i<N;i++) L_d[i+i*N] = 1;
		portfolio_d(N,Nmat,delta,Nopt,maturities,swaprates,L,L_d,v,v_d);
	    }

	    if( pass==4 ) {
		// plain evaluation for timing comparison //
		for(i=0;i<N;i++) {
		    L[i] = 0.05;
		}
		tac_portfolio(N, Nmat, delta, Nopt, maturities, swaprates, L, vptr);
	    }

	    if( pass==0 | pass==5 ) {
		//prepare for tac adjoint
		for(i=0;i<N;i++) L[i] = 0.05;
		for(i=0;i<N;i++) L_ad[i]=0;
		*vptr_ad = 1;
		tac_portfolio_ad(N,Nmat,delta,Nopt,maturities,swaprates,L, L_ad, vptr, vptr_ad);
	    }
	}
	    tt[pass] = ((double) clock())/CLOCKS_PER_SEC;
//	    if (pass>0) tt[pass-1] = (tt[pass]-tt[pass-1])*(1000.0/npath);
	    if (pass>0) tt[pass-1] = (tt[pass]-tt[pass-1])*(1000.0);

    }
    printf("\n--------------------------------------------------\n");
    printf("  timings in milliseconds per %d paths\n", npath);
    printf(" %10s  %10s  %10s  %10s  %10s\n", "portfolio(org)", "hc-b", "hc-f", "tac_portfolio", "ADM:tac_portfolio");
    printf(" %10.2f %10.2f %10.2f %10.2f %10.2f\n", tt[0], tt[1], tt[2], tt[3], tt[4]);
#endif//PORTFOLIO_ADM_TIMING


    //----------------------------------------------------------//
    //                                                          //
    // Next, do a full path + portfolio sensitivity             //
    // and a timing study                                       //
    //                                                          //
    // A real application would generate a different random     //
    // vector z for each path but here we set one and reuse it  //
    //                                                          //
    //----------------------------------------------------------//

    for (i=0; i<Nmat; i++) z[i]=0.3;
    for (pass=0; pass<6; pass++) {
	if (pass==0) {
	    npath = 1;
	}
	else if (pass<6) {
#ifdef NPATHTIMING
	    npath = NPATHTIMING;
#else
	    npath = 1000;
#endif
	}

	for (path=0; path<npath; path++) {

	    if (pass==0 | pass==1) {
		// plain evaluation for timing comparison //
		for(i=0;i<N;i++) {
		    L[i] = 0.05;
		    lambda[i] = 0.2;
		}
		path_calc(N,Nmat,delta,L,lambda,z);
		portfolio(N,Nmat,delta,Nopt,maturities,swaprates,L,v);
	    }
	    if (pass==0 | pass==2) {
		// hand-coded reverse mode //
		for(i=0;i<N;i++) {
		    L[i] = 0.05;
		    lambda[i] = 0.2;
		}
		path_calc_b1(N,Nmat,delta,L,lambda,z,L2,L3);
		portfolio_b(N,Nmat,delta,Nopt,maturities,swaprates,L,L_b,v);
		path_calc_b2(N,Nmat,delta,L,L_b,lambda,lambda_b,z,L2,L3);
	    }

	    if (pass==0 | pass==3) {
		// hand-coded forward mode //
		for(i=0;i<N;i++) L[i] = 0.05;
		for(i=0;i<N*N;i++) L_d[i] = 0;
		for(i=0;i<N;i++) L_d[i+i*N] = 1;
		path_calc_d(N,Nmat,delta,L,L_d,lambda,z);
		portfolio_d(N,Nmat,delta,Nopt,maturities,swaprates,L,L_d,v,v_d);
	    }

	    if( pass==0 | pass==4 ) {
		// plain evaluation for timing comparison //
		for(i=0;i<N;i++) {
		    L[i] = 0.05;
		    lambda[i] = 0.2;
		}
		model_ext(N, Nmat, delta, Nopt, maturities, z, swaprates, L, lambda, vptr);
	    }

	    if( pass==0 | pass==5 ) {
		//prepare for tac adjoint
		for(i=0;i<N;i++) {
		    L[i] = 0.05;
		    lambda[i] = 0.2;
		}
		for(i=0;i<N;i++) {
		    L_ad[i]=0;
		    lambda_ad[i]=0;
		}
		*vptr_ad = 1;
		model_ext_ad(N, Nmat, delta, Nopt, maturities, z, swaprates, L, L_ad, lambda, lambda_ad, vptr, vptr_ad);
	    }
	}

	if (pass==0) {

	    printf("\n==================================================\n");
	    printf("  path + portfolio sensitivity check, hand vs. tac(++)\n");
	    printf("\n==================================================\n");
	    printf( "%4sd %14s  %14s  %14s  \n",
		    "idx", "hand ADM", "TAC(++) ADM ", "rel. Diff");
	    for(i=0; i<N; i++) {
		printf("%4d %14.8f  %14.8f  %14.8f\n", i, L_b[i], L_ad[i], reldiff(L_b[i],L_ad[i], epsmach));
	    }

	}

	tt[pass] = ((double) clock())/CLOCKS_PER_SEC;
	if (pass>0) tt[pass-1] = (tt[pass]-tt[pass-1])*(1000.0/npath);
    }

    printf("\n==================================================\n");
    printf("          TIMINGS:\n");
    printf("==================================================\n");
    printf("  absolute timings in milliseconds per path (based on %d paths)\n", npath);
    printf(" val       : pathcalc+portfolio evaluation (Mikes original)\n");
    printf(" hc-b      : Mikes hand-coded reverse\n");
    printf(" hc-f      : Mikes hand-coded forward\n");
    printf(" tac++-val : pathcalc+portfolio evalutation (TAC++ reverse compliant version)\n");
    printf(" tac++-b   : TAC++ generated reverse\n");
    printf("-----------------------------------------------------------------------------\n");
    printf(" %10s  %10s  %10s  %10s  %10s\n", "val", "hc-b", "hc-f", "tac++-val", "tac++-b");
    printf(" %10.2f %10.2f %10.2f %10.2f %10.2f\n", tt[0], tt[1], tt[2], tt[3], tt[4]);
    printf("-----------------------------------------------------------------------------\n");
    printf("  relative timings\n");
    printf(" adjoint(hand)/func: %5.2f\n", tt[1]/tt[0]);
    printf(" tangent(hand)/func: %5.2f\n", tt[2]/tt[0]);
    printf(" adjoint(tac)/func : %5.2f\n", tt[4]/tt[3]);


    return 0;
}
