//======================================================================
//
//                    mike-codes.cpp
//
// Purpose:
//          - file contains all forward and derivative routines written
//            by Mike Giles (as found in file testlinadj.cpp) with
//            ADdouble replaced by double
//
//======================================================================
#include <math.h>

/* Monte Carlo LIBOR path calculation                      */
/* template enables use of AD through operator overloading */

//------------------------------
// MVNOTE: modifications are:
//         ADdouble --> double,
//         no longer template
//------------------------------
void path_calc(const int N, const int Nmat, const double delta,
               double L[], const double lambda[], const double z[])
{
    int      i, n;
    double   sqez, lam, con1;
    double v, vrat;

    for(n=0; n<Nmat; n++) {
	sqez = sqrt(delta)*z[n];

	v = 0.0;
	for (i=n+1; i<N; i++) {
	    lam  = lambda[i-n-1];
	    con1 = delta*lam;
	    v   += (con1*L[i])/(1.0+delta*L[i]);
	    vrat = exp(con1*v + lam*(sqez-0.5*con1));
	    L[i] = L[i]*vrat;
	}
    }
}

/* forward pathwise sensitivity calculation of deltas
   (commented out bits are for vega calculation) */

void path_calc_d(const int N, const int Nmat, const double delta,
                 double L[], double L_d[],
                 const double lambda[], const double z[])
{
    int    i, k, n, N2;
    double sqez, v, lam, faci, vrat, con1, con2, con3;

    double v1[N], v2[N], v_l[N];
//    vector<double> v1(N), v2(N), v_l(N);

    N2 = N*N;

    for(n=0; n<Nmat; n++) {
	sqez = sqrt(delta)*z[n];

	v = 0;
	for (k=0; k<N; k++) {
	    v1[k]=0;
	    v2[k]=0;
	}

	for (i=n+1; i<N; i++) {
	    lam  = lambda[i-n-1];
	    con1 = delta*lam;
	    faci = 1.0/(1.0+delta*L[i]);
	    v   += con1*L[i]*faci;
	    // v_l[i-n] = delta*L[i]*faci;
	    vrat = exp(con1*(-0.5*lam+v) + lam*sqez);

	    con2 = con1*faci*faci;
	    con3 = con1*L[i]*vrat;

	    for (k=0; k<i+1; k++) {
		v1[k]        += con2*L_d[k+i*N];
		L_d[k+i*N]    = L_d[k+i*N]*vrat + con3*v1[k];
		// v2[k]        += con2*L_d[k+i*N+N2];
		// L_d[k+i*N+N2] = L_d[k+i*N+N2]*vrat + con3*v2[k];
	    }

	    /*
	      for (k=0;k<=i-n-1; k++) {
	      L_d[k+i*N+N2] += con3*v_l[k];
	      }
	      L_d[(i-n-1)+i*N+N2] += L[i]*vrat*((v-lam)*delta+sqez);
	    */

	    L[i] = L[i]*vrat;
	}
    }
}


/* forward path calculation storing data
   for subsequent reverse path calculation */

void path_calc_b1(const int N, const int Nmat, const double delta,
                  double L[], const double lambda[], const double z[],
                  double L2[], double L3[])
{
    int    i, n;
    double sqez, lam, v, vrat, con1;

    for (i=0; i<N; i++) L2[i] = L[i];
   
    for(n=0; n<Nmat; n++) {
	sqez = sqrt(delta)*z[n];

	v = 0;
	for (i=n+1; i<N; i++) {
	    lam  = lambda[i-n-1];
	    con1 = delta*lam;
	    v   += (con1*L[i])/(1.0+delta*L[i]);
	    vrat = exp(con1*(-0.5*lam+v) + lam*sqez);
	    L[i] = L[i]*vrat;

	    // store these values for reverse path //
	    L2[i+(n+1)*N] = L[i];
	    L3[i+(n+1)*N] = L[i]*((v-lam)*delta+sqez);
	}
    }
}


/* reverse path calculation of deltas using stored data
   (also calculates vegas at minimal additional cost)  */

void path_calc_b2(const int N, const int Nmat, const double delta,
                  const double L[], double L_b[],
                  const double lambda[], double lambda_b[],
                  const double z[], const double L2[], const double L3[])
{
    int    i, n;
    double faci;
    double v1[N];
    // vector<double> v1(N);

    for (i=0; i<N; i++) lambda_b[i] = 0;

    for (n=Nmat-1; n>=0; n--) {
	v1[N-1] = 0;
	for (i=N-1; i>n; i--) 
	    v1[i-1] = v1[i] + lambda[i-n-1]*L2[i+(n+1)*N]*L_b[i];

	for (i=n+1; i<N; i++) {
	    faci = 1.0/(1.0+delta*L2[i+n*N]);

	    lambda_b[i-n-1] += v1[i-1] * delta*delta*L2[i+n*N]*faci
		+ L3[i+(n+1)*N]*L_b[i];

	    L_b[i] = L_b[i]*L2[i+(n+1)*N]/L2[i+n*N]
		+ v1[i-1]*(delta*delta*lambda[i-n-1])*faci*faci;
	}
    }
}


/* calculate the portfolio value v */
/* template enables use of AD through operator overloading */

//------------------------------
// MVNOTE: modifications:
//         ADdouble --> double,
//         no longer template
//         vector<double> B(N) --> B[N]
//         vector<double> S(N) --> S[N]
//------------------------------
void portfolio(const int N, const int Nmat, const double delta,
               const int Nopt, const int maturities[],
               const double swaprates[],
               const double L[], double* v )
{
    int      i, m, n;
    double b, s, swapval;
//    vector<double> B(N), S(N);
    double B[N], S[N];

    b = 1.0;
    s = 0.0;

    for (n=Nmat; n<N; n++) {
	b    = b/(1.0+delta*L[n]);
	s    = s + delta*b;
	B[n] = b;
	S[n] = s;
    }

    *v = 0;

    for (i=0; i<Nopt; i++){
	m = maturities[i] + Nmat-1;
	swapval = B[m] + swaprates[i]*S[m] - 1.0;
	if (swapval<0)
	    *v += -100.0*swapval;
    }

    // apply discount //

    for (n=0; n<Nmat; n++)
	*v = *v/(1.0+delta*L[n]);
}


/* calculate the portfolio value v, and its sensitivity to L */
/* hand-code forward mode sensitivity */


void portfolio_d(const int N, const int Nmat, const double delta,
                 const int Nopt, const int maturities[],
                 const double swaprates[],
                 const double L[], const double L_d[],
                 double* v, double v_d[])
{
    int      i, k, m, n;
    double   b, s, swapval, swapval_d;
    double   B[N], S[N], B_d[N*N], S_d[N*N], b_d[N], s_d[N];
    // vector<double> B(N), S(N), B_d(N*N), S_d(N*N), b_d(N), s_d(N);

    b = 1.0;
    s = 0.0;
    for (k=0; k<N; k++) {
	b_d[k] = 0;
	s_d[k] = 0;
    }

    for (n=Nmat; n<N; n++) {
	b    = b/(1.0+delta*L[n]);
	s    = s + delta*b;
	B[n] = b;
	S[n] = s;

	for (k=0; k<N; k++) {
	    b_d[k] = (b_d[k]-delta*b*L_d[k+n*N])/(1.0+delta*L[n]);
	    s_d[k] = s_d[k] + delta*b_d[k];
	    B_d[k+n*N] = b_d[k];
	    S_d[k+n*N] = s_d[k];
	}
    }

    *v = 0;
    for (k=0; k<N; k++)
	v_d[k] = 0;

    for (i=0; i<Nopt; i++){
	m = maturities[i] + Nmat-1;
	swapval = B[m] + swaprates[i]*S[m] - 1.0;
	if (-swapval>0.0) {
	    *v += -100.0*swapval;
	    for (k=0; k<N; k++) {
		swapval_d = B_d[k+m*N] + swaprates[i]*S_d[k+m*N];
		v_d[k] += -100.0*swapval_d;
	    }
	}
    }


    // apply discount //

    b = 1.0;
    for (n=0; n<Nmat; n++)
	b = b/(1.0+delta*L[n]);

    *v = b* *v;

    for (k=0; k<N; k++) v_d[k] = b*v_d[k];

    for (i=0; i<Nmat; i++) {
	for (k=0; k<N; k++)
	    v_d[k] += -*v*delta*L_d[k+i*N]/(1.0+delta*L[i]);
    }
}



/* calculate the portfolio value v, and its sensitivity to L */
/* hand-code reverse mode sensitivity */

void portfolio_b(const int N, const int Nmat, const double delta, 
                 const int Nopt, const int maturities[],
                 const double swaprates[],
                 const double L[], double L_b[], double* v) 
{
    int      m, n;
    double   b, s, swapval;
    double   B[N], S[N], B_b[N], S_b[N];
    // vector<double> B(N), S(N), B_b(N), S_b(N);

    for (n=0; n<N; n++) L_b[n] = 0;

    b = 1.0;
    s = 0.0;
    for (n=Nmat; n<N; n++) {
	b    = b/(1.0+delta*L[n]);
	s    = s + delta*b;
	B[n] = b;
	S[n] = s;
    }

    *v = 0;

    for (n=0; n<N; n++) {
	B_b[n] = 0;
	S_b[n] = 0;
    }

    for (n=0; n<Nopt; n++){
	m = maturities[n] + Nmat-1;
	swapval = B[m] + swaprates[n]*S[m] - 1.0;
	if (swapval<0) {
	    *v      += -100*swapval;
	    S_b[m] += -100*swaprates[n];
	    B_b[m] += -100;
	}
    }

    for (n=N-1; n>Nmat-1; n--) {
	B_b[n]   += delta*S_b[n];
	S_b[n-1] += S_b[n];
    }

    for (n=N-1; n>Nmat-1; n--) {
	L_b[n]    = -B_b[n]*B[n]*delta/(1.0+delta*L[n]);
	if (n>0)
	    B_b[n-1] +=  B_b[n]/(1.+delta*L[n]);
    }

    // apply discount //

    b = 1.0;
    for (n=0; n<Nmat; n++) b = b/(1.0+delta*L[n]);

    *v = b* *v;

    for (n=0; n<Nmat; n++){
	L_b[n] = -*v*delta/(1.0+delta*L[n]);
    }

    for (n=Nmat; n<N; n++){
	L_b[n] = b*L_b[n];
    }
}

