//======================================================================
//
//                    modeldriver.c
//
// Purpose:
//          - compare results of forward code (as modified to be TAC++
//            compliant) and forward code written by Mike Giles
//            (with ADdouble replaced by double)
//          - perform "cost2" test of toplevel function
//            (i.e. the function that is going to be differentiated)
// 
//======================================================================


/* Program to compute deltas and vegas of swaption portfolio
   from forward and reverse mode pathwise sensitivities 
   written by Mike Giles in 2005-7 based on code written by
   Zhao and Glasserman at Columbia University */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>


#include "tac++-codes.h"
#include "mikes-codes.h"

/* -------------------------------------------------------- */

int main(){

    double delta = 0.25; /* LIBOR interval  */

    // data for swaption portfolio //
    int    Nopt = 15;
    int    maturities[] = {4,4,4,8,8,8,20,20,20,28,28,28,40,40,40};
    double swaprates[]  = {.045,.05,.055,.045,.05,.055,.045,.05,
			   .055,.045,.05,.055,.045,.05,.055 };

    int i, npath, pass, path;
    int       N, Nmat;
    double    *v, *v_d, *L, *L_b, *L_d, *L2, *L3, 
	*lambda, *lambda_b, *z, tt[6];

    Nmat = 40;
    N = Nmat+40;

    v   = (double*) malloc(sizeof(double));
    L   = (double*) malloc(N*sizeof(double));
    L_b = (double*) malloc(N*sizeof(double));
    v_d = (double*) malloc(2*N*sizeof(double));
    L_d = (double*) malloc(2*N*N*sizeof(double));
    L2  = (double*) malloc(N*N*sizeof(double));
    L3  = (double*) malloc(N*N*sizeof(double));

    lambda   = (double*) malloc(N*sizeof(double));
    lambda_b = (double*) malloc(N*sizeof(double));
    z        = (double*) malloc(N*sizeof(double));

    //------------------------------
    // variables required for running
    // tac(++) modifications
    //
    // MVCOMMENT:
    // - C++ reference type variables
    //   are not yet supported by tac(++).
    //   Thus, we must introduce pointer
    //   type variables for "output" arugments
    //   (2 vars are required for cost2 check)
    //------------------------------
    double *vptr;  
    double *vptr2;
    double *vextptr;
    double *vextptr2;
    vptr     = (double*) malloc(sizeof(double));
    vptr2    = (double*) malloc(sizeof(double));
    vextptr  = (double*) malloc(sizeof(double));
    vextptr2 = (double*) malloc(sizeof(double));

    //============================================================
    //
    //          C O M P A R E
    //
    //          C O S T 2   V E R I F I C A T I O N
    //
    // Note:
    //       - cost2 test runs the toplevel function twice with the
    //         same arguments and compares the function results.
    //       - we thus check, that there are no "sideeffects" on
    //         input arguments.
    //
    //============================================================
    //------------------------------
    // Mikes original full pathcalc+portfolio evaluation
    //
    // MVNOTE: must be done last, as
    //------------------------------
    for(i=0; i<Nmat; i++) z[i]=0.3;
    for(i=0; i<N; i++) {
	L[i] = 0.05;
	lambda[i] = 0.2;
    }
    path_calc(N,Nmat,delta,L,lambda,z);
    portfolio(N,Nmat,delta,Nopt,maturities,swaprates,L,v); 


    //------------------------------
    // MVNOTE:: L must be reinitialised,
    //          as path_calc has modified it!
    //------------------------------
    for(i=0; i<Nmat; i++) z[i]=0.3;
    for(i=0; i<N; i++) {
	L[i] = 0.05;
	lambda[i] = 0.2;
    }

    //------------------------------
    // 2 consecutive calls of toplevel
    // function "model"
    // (i.e. full pathcalc+portfolio calcaluation)
    //------------------------------
    model(N, Nmat, delta, Nopt, maturities, z, swaprates, L, lambda, vptr);
    model(N, Nmat, delta, Nopt, maturities, z, swaprates, L, lambda, vptr2);

    //------------------------------
    // 2 consecutive calls of toplevel
    // function "model_ext"
    // (i.e. full pathcalc+portfolio evaluation)
    // with version of path_calc that
    // is TAC++ compliant in reverse mode.
    //------------------------------
    model_ext(N, Nmat, delta, Nopt, maturities, z, swaprates, L, lambda, vextptr);
    model_ext(N, Nmat, delta, Nopt, maturities, z, swaprates, L, lambda, vextptr2);

    printf("\n==================================================\n");
    printf("          COMPARE: path + portfolio function value check,\n");
    printf("                   original vs. tac++-compliant:\n");
    printf("==================================================\n");
    printf("  model (Mikes version)       : f    = %12.8f\n", *v);
    printf("  model (tac++-compliant)     : f    = %12.8f\n", *vptr);
    printf("  difference                  : d    = %12.8f\n", reldiff(*v, *vptr, epsmach));

    printf("\n==================================================\n");
    printf("          COMPARE: path + portfolio function value check,\n");
    printf("                   original vs. tac++-compliant for reverse mode:\n");
    printf("==================================================\n");
    printf("  model (Mikes version)       : f    = %12.8f\n", *v);
    printf("  model_ext (tac++-compliant) : f    = %12.8f\n", *vextptr);
    printf("  difference                  : d    = %12.8f\n", reldiff(*v, *vextptr, epsmach));


    printf("\n==================================================\n");
    printf("          COST2 check:\n");
    printf("==================================================\n");
    printf("  model,     first evaluation : f1   = %12.8f\n", *vptr);
    printf("  model,     second evaluation: f2   = %12.8f\n", *vptr2);
    printf("  model,     difference       : diff = %12.8f\n", reldiff(*vptr, *vptr2, epsmach));
    printf("\n");
    printf("  model_ext, first evaluation : f1   = %12.8f\n", *vextptr);
    printf("  model_ext, second evaluation: f2   = %12.8f\n", *vextptr2);
    printf("  model_ext, difference       : diff = %12.8f\n", reldiff(*vextptr, *vextptr2, epsmach));




    //============================================================
    //
    //          T I M I N G
    //
    //============================================================

    //----------------------------------------------------------//
    //                                                          //
    // Next, do a full path + portfolio sensitivity             //
    // and a timing study                                       //
    //                                                          //
    // A real application would generate a different random     //
    // vector z for each path but here we set one and reuse it  //
    //                                                          //
    //----------------------------------------------------------//
    for(i=0; i<Nmat; i++) z[i]=0.3;
    for (pass=0; pass<4; pass++) {
	if (pass==0) {
	    npath = 1;
	}
	else if (pass<4) {
#ifdef NPATHTIMING
	    npath = NPATHTIMING;
#else
	    npath = 1000;
#endif
	}

	for (path=0; path<npath; path++) {

	    if (pass == 0 | pass==1) {
		// plain evaluation for timing comparison
		for(i=0;i<N;i++) {
		    L[i] = 0.05;
		    lambda[i] = 0.2;
		}
		path_calc(N,Nmat,delta,L,lambda,z);
		portfolio(N,Nmat,delta,Nopt,maturities,swaprates,L,v);
	    }

	    if (pass==2) {
		// model evaluation
		for(i=0;i<N;i++) {
		    L[i] = 0.05;
		    lambda[i] = 0.2;
		}
		model(N, Nmat, delta, Nopt, maturities, z, swaprates, L, lambda, vptr);
	    }

	    if (pass==3) {
		// model evaluation (path_calc "extended" for adjoint code generation)
		for(i=0;i<N;i++) {
		    L[i] = 0.05;
		    lambda[i] = 0.2;
		}
		model_ext(N, Nmat, delta, Nopt, maturities, z, swaprates, L, lambda, vptr2);
	    }

	}

	tt[pass] = ((double) clock())/CLOCKS_PER_SEC;
	if (pass>0) tt[pass-1] = (tt[pass]-tt[pass-1])*(1000.0/npath);
    }

    printf("\n==================================================\n");
    printf("          TIMINGS:\n");
    printf("==================================================\n");
    printf("  timings are in milliseconds per path, based on npath=%d pathes:\n", npath);
    printf(" %12s  %12s  %12s\n", "Original", "model", "model_ext");
    printf(" %12.2f %12.2f %12.2f \n", tt[0], tt[1], tt[2]);

    return 0;
}
