//======================================================================
//
//                    tac++-codes.c
//
// Purpose:
//          - file contains TAC++ compliant versions of routines
//            portfolio and path_calc (that are define in testlinadj.cpp)
//
//          - file also contains
//
// NOTE:
//       - below in the code we use mathematical intrinsic functions of
//         C (i.e. exp, sqrt, fabs)
//       - As we will later compile this file (or the generated derivative)
//         with a C++ compiler we must include <cmath> which cannot be
//         handled by TAC++.
//       - Our actual workaround is to include <cmath> by postprocessing
//         this (or a generated file) within our Makefile.
// 
//======================================================================

#include <math.h>

//------------------------------
// MVCOMMENT: MODEL_EXT_INLINED: (reverse mode only!)
// path_calc_ext code is directly "inlined" into toplevel
// routine model_ext rather than calling path_calc_ext
//------------------------------
#define MODEL_EXT_INLINED

#define min(a,b) ((a) <= (b) ? (a) : (b))


//--------------------------------------------------
//          utility routine
//          used in driver to compare results
//--------------------------------------------------
double reldiff( double a, double b, const double epsmach) {
    double rdiff;
    double amax;
    double adiff;
    adiff = fabs(a-b);
    if( fabs(a) >= fabs(b) ) {
	amax = fabs(a);
    }
    else {
	amax = fabs(b);
    }
    if ( amax < epsmach )
	rdiff = adiff * 1.e4;
    else {
	rdiff = adiff / amax;
    }
    return rdiff;
}


//============================================================
//
//                    tac_path_calc
//
// Note:
// - this is the modified path_calc routine, that we do
//   differentiate in forward mode (scalar and vector)
//
// - method is not based on template ADouble but uses simple double
// - input argument "L" is not modified, instead the modified
//   path is returned via output argument "Lout"
//
//============================================================
void tac_path_calc(const int N, const int Nmat, const double delta,
		   double L[], double lambda[], const double z[], double Lout[])
{
    int      i, n;
    double   sqez, lam, con1;
    double   v, vrat;

    // copy L ---> Lout
    for(i=0; i<N; i++)
	Lout[i] = L[i];

    // compute path,
    // based on Lout
    for(n=0; n<Nmat; n++) {
	sqez = sqrt(delta)*z[n];

	v = 0.0;
	for (i=n+1; i<N; i++) {
	    lam  = lambda[i-n-1];
	    con1 = delta*lam;
	    v   += (con1*Lout[i])/(1.0+delta*Lout[i]);
	    vrat = exp(con1*v + lam*(sqez-0.5*con1));
	    Lout[i] = Lout[i]*vrat;
	}
    }

}


//============================================================
//
//                    tac_path_calc_ext
//
// Note:
// - this is the modified path_calc routine, that we do
//   differentiate in reverse mode.
//
// - method is a reimplementation of tac_path_calc, but
//   extends some variables by one dimension, which allows
//   for TAC++ to treat the essential loop as "non-destructive"
//
// - code contains TAC++ loop directives for more efficient
//   reverse code.
//
//============================================================
void tac_path_calc_ext(const int N, const int Nmat, const double delta,
		       double L[], double lambda[], const double z[], double Lout[])
{
    int      i, n;
    double   sqez, lam, con1;
    double   v, vrat;

    int j,k;
    double Lext[Nmat+1][N], vext[Nmat][N], vratext[Nmat][N];


#pragma tac loop = parallel
    for(j=0; j<N; j++) {
	Lext[0][j] = L[j];
#pragma tac loop = parallel
	for(k=0; k<Nmat; k++) {
	    Lext[k+1][j] = L[j];
	    vext[k][j] = 0.0;
	}
    }

#pragma tac loop = nondestructive  //MVNOTE::causes erroneous derivatives in pure-mode
    for(n=0; n<Nmat; n++) {
	sqez = sqrt(delta)*z[n];

#pragma tac loop = nondestructive
	for (i=n+1; i<N; i++) {
	    lam  = lambda[i-n-1];
	    con1 = delta*lam;
	    vext[n][i]    = vext[n][i-1] + (con1*Lext[n][i])/(1.0+delta*Lext[n][i]);
	    vratext[n][i] = exp(con1*vext[n][i] + lam*(sqez-0.5*con1));
	}

#pragma tac loop = parallel
	for(i=n+1;i<N;i++) {
	    Lext[n+1][i] = Lext[n][i]*vratext[n][i];
	}

    }

    Lout[0] = L[0];

#pragma tac loop = parallel
    for(j=1; j<N; j++) {
	Lout[j] = Lext[min(j,Nmat)][j];
    }
}


//============================================================
//
//                    tac_portfolio
//
// Note:
//
// - this is a reimplementation of Mikes original portfolio routine,
//   modified by replacing C++ by C language elements, as there are:
//   - no more template-function
//   - Adouble --> double
//   - vector<Adouble> B(N) --> double B[N]
//
// - one modification was made w.r.t. TAC++ bug in
//   recomputation of required variables:
//   loop index of last loop was changed from "i" to "p"
//
// - added tac++ loop directives
//   (required for tac++ reverse mode code generation)
//
//============================================================
void tac_portfolio(const int N, const int Nmat, const double delta,
		   const int Nopt, const int maturities[],
		   const double swaprates[],
		   double L[], double* v )
{
    int      i, m, n;

    //------------------------------
    // MV::tac++ problem
    // use additional loop variable in last loop
    // otherwise we get problems in recomputation
    //------------------------------
    int p;
    double b, s, swapval;

    double B[N], S[N];

    b = 1.0;
    s = 0.0;

    for (n=Nmat; n<N; n++) {
	b    = b/(1.0+delta*L[n]);
	s    = s + delta*b;
	B[n] = b;
	S[n] = s;
    }

    *v = 0;

#pragma tac loop = nondestructive
    for (i=0; i<Nopt; i++){
	m = maturities[i] + Nmat-1;
	swapval = B[m] + swaprates[i]*S[m] - 1.0;
	if (swapval<0)
	    *v += -100.0*swapval;
    }

    // apply discount //

    for (p=0; p<Nmat; p++) //MVNOTE::using "i" yields problems in TAC++ reverse mode
	*v = *v/(1.0+delta*L[p]);
}


//============================================================
//
//                    tac_portfolio_ext
//
// Note:
//
// - this is a modified version of the tac_portfolio routine
//   with modifications such that TAC++ is able to generate a
//   more efficient adjoint code
//
// - main modifications are done within two loop kernels,
//   thus TAC++ can treat this loops as "non-destructive" and
//   so avoids costly recomputations.
//
//============================================================
void tac_portfolio_ext(const int N, const int Nmat, const double delta,
		       const int Nopt, const int maturities[],
		       const double swaprates[],
		       double L[], double* v )
{
    int      i, m, n;

    //------------------------------
    // MV::tac++ problem
    // use additional loop variable in last loop
    // otherwise we get problems in recomputation
    //------------------------------
    int p;
    double swapval;

    double B[N], S[N];

    double vtmp[Nmat];

    B[Nmat-1] = 1.0;
    S[Nmat-1] = 0.0;
#pragma tac loop = nondestructive
    for (n=Nmat; n<N; n++) {
	B[n]    = B[n-1]/(1.0+delta*L[n]);
	S[n]    = S[n-1] + delta*B[n];
    }

    vtmp[0] = 0;
#pragma tac loop = nondestructive
    for (i=0; i<Nopt; i++){
	m = maturities[i] + Nmat-1;
	swapval = B[m] + swaprates[i]*S[m] - 1.0;
	if (swapval<0)
	    vtmp[0] += -100.0*swapval;
    }

    // apply discount //
    vtmp[0] = vtmp[0]/(1.0+delta*L[0]);

#pragma tac loop = nondestructive
    for (p=1; p<Nmat; p++)
	vtmp[p] = vtmp[p-1]/(1.0+delta*L[p]);

    *v = vtmp[Nmat-1];
}


//============================================================
//
//                    model
//
// Note:
// - implements Mike's "full path + portfolio evaluation",
//   i.e. what we call "cost function" (L,lambda) ---> v
//
// - this is the toplevel function that will be differentiated
//   by TAC++ in forward mode.
//
//============================================================
void model(const int N, const int Nmat, const double delta,
	   const int Nopt, const int maturities[],
	   const double z[], const double swaprates[],
	   double L[], double lambda[], double *v )
{
    double Lout[N];
    tac_path_calc(N, Nmat, delta, L, lambda, z, Lout);

    tac_portfolio(N, Nmat, delta, Nopt, maturities, swaprates, Lout, v);
}


//============================================================
//
//                    model_ext
//
// Note:
// - implements Mike's "full path + portfolio evaluation",
//   i.e. what we call "cost function" (L,lambda) ---> v
//
// - this is the toplevel function that will be differentiated
//   by TAC++ in reverse mode.
//
// - uses modified ("extended") version of path_calc, such
//   that TAC++ is able to generate more efficient adjoint code.
//
//============================================================
void model_ext(const int N, const int Nmat, const double delta,
	       const int Nopt, const int maturities[],
	       const double z[], const double swaprates[],
	       double L[], double lambda[], double *v )
{
    double Lout[N];

//------------------------------
// MVINFO:
// - switch below enables us to
//   have the path_calc part directly
//   in model rather than calling
//   tac_path_calc_ext
//------------------------------
#ifndef MODEL_EXT_INLINED
    tac_path_calc_ext(N, Nmat, delta, L, lambda, z, Lout);
#else
    int      i, n;
    double   sqez, lam, con1;
    double   vrat;

    int j,k;
    double Lext[Nmat+1][N], vext[Nmat][N], vratext[Nmat][N];


#pragma tac loop = parallel
    for(j=0; j<N; j++) {
	Lext[0][j] = L[j];
#pragma tac loop = parallel
	for(k=0; k<Nmat; k++) {
	    Lext[k+1][j] = L[j];
	    vext[k][j] = 0.0;
	}
    }

//------------------------------
//MV:: AAAAA:causes wrong derivatives
//     in pure-reverse mode
//------------------------------
#pragma tac loop = nondestructive
    for(n=0; n<Nmat; n++) {
	sqez = sqrt(delta)*z[n];

#pragma tac loop = nondestructive
	for (i=n+1; i<N; i++) {
	    lam  = lambda[i-n-1];
	    con1 = delta*lam;
	    vext[n][i]    = vext[n][i-1] + (con1*Lext[n][i])/(1.0+delta*Lext[n][i]);
	    vratext[n][i] = exp(con1*vext[n][i] + lam*(sqez-0.5*con1));
	}

#pragma tac loop = parallel
	for(i=n+1;i<N;i++) {
	    Lext[n+1][i] = Lext[n][i]*vratext[n][i];
	}

    }

    Lout[0] = L[0];
#pragma tac loop = parallel
    for(j=1; j<N; j++) {
	Lout[j] = Lext[min(j,Nmat)][j];
    }

#endif

    tac_portfolio_ext(N, Nmat, delta, Nopt, maturities, swaprates, Lout, v);
}


