
/* Program to compute deltas and vegas of swaption portfolio
   from forward and reverse mode pathwise sensitivities 
   written by Mike Giles in 2005-7 based on code written by
   Zhao and Glasserman at Columbia University */

#include "fadbad.h"
#include "fadiff.h"
#include "badiff.h"

using namespace std;
using namespace fadbad;

#include <stdlib.h>
#include <stdio.h>
#include <ctime>
#include <cmath>


/* Monte Carlo LIBOR path calculation                      */
/* template enables use of AD through operator overloading */

template <typename ADdouble>
void path_calc(const int N, const int Nmat, const double delta,
               ADdouble L[], const double lambda[], const double z[])
{
  int      i, n;
  double   sqez, lam, con1;
  ADdouble v, vrat;

  for(n=0; n<Nmat; n++) {
    sqez = sqrt(delta)*z[n];

    v = 0.0;
    for (i=n+1; i<N; i++) {
      lam  = lambda[i-n-1];
      con1 = delta*lam;
      v   += (con1*L[i])/(1.0+delta*L[i]);
      vrat = exp(con1*v + lam*(sqez-0.5*con1));
      L[i] = L[i]*vrat;
    }
  }
}

/* forward pathwise sensitivity calculation of deltas
   (commented out bits are for vega calculation) */

void path_calc_d(const int N, const int Nmat, const double delta,
                 double L[], double L_d[],
                 const double lambda[], const double z[])
{
  int    i, k, n, N2;
  double sqez, v, lam, faci, vrat, con1, con2, con3;

  double v1[N], v2[N], v_l[N];
  // vector<double> v1(N), v2(N), v_l(N);

  N2 = N*N;

  for(n=0; n<Nmat; n++) {
    sqez = sqrt(delta)*z[n];

    v = 0;
    for (k=0; k<N; k++) {
      v1[k]=0;
      v2[k]=0;
    }

    for (i=n+1; i<N; i++) {
      lam  = lambda[i-n-1];
      con1 = delta*lam;
      faci = 1.0/(1.0+delta*L[i]);
      v   += con1*L[i]*faci;
      // v_l[i-n] = delta*L[i]*faci;
      vrat = exp(con1*(-0.5*lam+v) + lam*sqez);

      con2 = con1*faci*faci;
      con3 = con1*L[i]*vrat;

      for (k=0; k<i+1; k++) {
        v1[k]        += con2*L_d[k+i*N];
        L_d[k+i*N]    = L_d[k+i*N]*vrat + con3*v1[k];
        // v2[k]        += con2*L_d[k+i*N+N2];
        // L_d[k+i*N+N2] = L_d[k+i*N+N2]*vrat + con3*v2[k];
      }

      /*
      for (k=0;k<=i-n-1; k++) {
        L_d[k+i*N+N2] += con3*v_l[k];
      }
      L_d[(i-n-1)+i*N+N2] += L[i]*vrat*((v-lam)*delta+sqez);
      */

      L[i] = L[i]*vrat;
    }
  }
}


/* forward path calculation storing data
   for subsequent reverse path calculation */

void path_calc_b1(const int N, const int Nmat, const double delta,
                  double L[], const double lambda[], const double z[],
                  double L2[], double L3[])
{
  int    i, n;
  double sqez, lam, v, vrat, con1;

  for (i=0; i<N; i++) L2[i] = L[i];
   
  for(n=0; n<Nmat; n++) {
    sqez = sqrt(delta)*z[n];

    v = 0;
    for (i=n+1; i<N; i++) {
      lam  = lambda[i-n-1];
      con1 = delta*lam;
      v   += (con1*L[i])/(1.0+delta*L[i]);
      vrat = exp(con1*(-0.5*lam+v) + lam*sqez);
      L[i] = L[i]*vrat;

      // store these values for reverse path //
      L2[i+(n+1)*N] = L[i];
      L3[i+(n+1)*N] = L[i]*((v-lam)*delta+sqez);
    }
  }
}


/* reverse path calculation of deltas using stored data
   (also calculates vegas at minimal additional cost)  */

void path_calc_b2(const int N, const int Nmat, const double delta,
                  const double L[], double L_b[],
                  const double lambda[], double lambda_b[],
                  const double z[], const double L2[], const double L3[])
{
  int    i, n;
  double faci;
  double v1[N];
  // vector<double> v1(N);

  for (i=0; i<N; i++) lambda_b[i] = 0;

  for (n=Nmat-1; n>=0; n--) {
    v1[N-1] = 0;
    for (i=N-1; i>n; i--) 
      v1[i-1] = v1[i] + lambda[i-n-1]*L2[i+(n+1)*N]*L_b[i];

    for (i=n+1; i<N; i++) {
      faci = 1.0/(1.0+delta*L2[i+n*N]);

      lambda_b[i-n-1] += v1[i-1] * delta*delta*L2[i+n*N]*faci
                         + L3[i+(n+1)*N]*L_b[i];

      L_b[i] = L_b[i]*L2[i+(n+1)*N]/L2[i+n*N]
              + v1[i-1]*(delta*delta*lambda[i-n-1])*faci*faci;
    }
  }
}


/* calculate the portfolio value v */
/* template enables use of AD through operator overloading */

template <typename ADdouble>
void portfolio(const int N, const int Nmat, const double delta,
               const int Nopt, const int maturities[],
               const double swaprates[],
               const ADdouble L[], ADdouble& v )
{
  int      i, m, n;
  ADdouble b, s, swapval;
  vector<ADdouble> B(N), S(N);

  b = 1.0;
  s = 0.0;

  for (n=Nmat; n<N; n++) {
    b    = b/(1.0+delta*L[n]);
    s    = s + delta*b;
    B[n] = b;
    S[n] = s;
  }

  v = 0;

  for (i=0; i<Nopt; i++){
    m = maturities[i] + Nmat-1;
    swapval = B[m] + swaprates[i]*S[m] - 1.0;
    if (swapval<0)
      v += -100.0*swapval;
  }

  // apply discount //

  for (n=0; n<Nmat; n++)
    v = v/(1.0+delta*L[n]);
}


/* calculate the portfolio value v, and its sensitivity to L */
/* hand-code forward mode sensitivity */


void portfolio_d(const int N, const int Nmat, const double delta,
                 const int Nopt, const int maturities[],
                 const double swaprates[],
                 const double L[], const double L_d[],
                 double& v, double v_d[])
{
  int      i, k, m, n;
  double   b, s, swapval, swapval_d;
  double   B[N], S[N], B_d[N*N], S_d[N*N], b_d[N], s_d[N];
  // vector<double> B(N), S(N), B_d(N*N), S_d(N*N), b_d(N), s_d(N);

  b = 1.0;
  s = 0.0;
  for (k=0; k<N; k++) {
    b_d[k] = 0;
    s_d[k] = 0;
  }

  for (n=Nmat; n<N; n++) {
    b    = b/(1.0+delta*L[n]);
    s    = s + delta*b;
    B[n] = b;
    S[n] = s;

    for (k=0; k<N; k++) {
      b_d[k] = (b_d[k]-delta*b*L_d[k+n*N])/(1.0+delta*L[n]);
      s_d[k] = s_d[k] + delta*b_d[k];
      B_d[k+n*N] = b_d[k];
      S_d[k+n*N] = s_d[k];
    }
  }

  v = 0;
  for (k=0; k<N; k++)
    v_d[k] = 0;

  for (i=0; i<Nopt; i++){
    m = maturities[i] + Nmat-1;
    swapval = B[m] + swaprates[i]*S[m] - 1.0;
    if (-swapval>0.0) {
      v += -100.0*swapval;
      for (k=0; k<N; k++) {
        swapval_d = B_d[k+m*N] + swaprates[i]*S_d[k+m*N];
        v_d[k] += -100.0*swapval_d;
      }
    }
  }


  // apply discount //

  b = 1.0;
  for (n=0; n<Nmat; n++)
    b = b/(1.0+delta*L[n]);

  v = b*v;

  for (k=0; k<N; k++) v_d[k] = b*v_d[k];

  for (i=0; i<Nmat; i++) {
    for (k=0; k<N; k++)
      v_d[k] += -v*delta*L_d[k+i*N]/(1.0+delta*L[i]);
  }
}



/* calculate the portfolio value v, and its sensitivity to L */
/* hand-code reverse mode sensitivity */

void portfolio_b(const int N, const int Nmat, const double delta, 
                 const int Nopt, const int maturities[],
                 const double swaprates[],
                 const double L[], double L_b[], double& v) 
{
  int      m, n;
  double   b, s, swapval;
  double   B[N], S[N], B_b[N], S_b[N];
  // vector<double> B(N), S(N), B_b(N), S_b(N);

  for (n=0; n<N; n++) L_b[n] = 0;

  b = 1.0;
  s = 0.0;
  for (n=Nmat; n<N; n++) {
    b    = b/(1.0+delta*L[n]);
    s    = s + delta*b;
    B[n] = b;
    S[n] = s;
  }

  v = 0;

  for (n=0; n<N; n++) {
    B_b[n] = 0;
    S_b[n] = 0;
  }

  for (n=0; n<Nopt; n++){
    m = maturities[n] + Nmat-1;
    swapval = B[m] + swaprates[n]*S[m] - 1.0;
    if (swapval<0) {
      v      += -100*swapval;
      S_b[m] += -100*swaprates[n];
      B_b[m] += -100;
    }
  }

  for (n=N-1; n>Nmat-1; n--) {
    B_b[n]   += delta*S_b[n];
    S_b[n-1] += S_b[n];
  }

  for (n=N-1; n>Nmat-1; n--) {
    L_b[n]    = -B_b[n]*B[n]*delta/(1.0+delta*L[n]);
    if (n>0)
      B_b[n-1] +=  B_b[n]/(1.+delta*L[n]);
  }

  // apply discount //

  b = 1.0;
  for (n=0; n<Nmat; n++) b = b/(1.0+delta*L[n]);

  v = b*v;

  for (n=0; n<Nmat; n++){
    L_b[n] = -v*delta/(1.0+delta*L[n]);
  }

  for (n=Nmat; n<N; n++){
    L_b[n] = b*L_b[n];
  }
}

/* -------------------------------------------------------- */

int main(){

  double delta = 0.25; /* LIBOR interval  */

  // data for swaption portfolio //
  int    Nopt = 15;
  int    maturities[] = {4,4,4,8,8,8,20,20,20,28,28,28,40,40,40};
  double swaprates[]  = {.045,.05,.055,.045,.05,.055,.045,.05,
                         .055,.045,.05,.055,.045,.05,.055 };

  int       i, N, Nmat, pass, path, npath;
  double    v, *v_d, *L, *L_b, *L_d, *L2, *L3, 
            *lambda, *lambda_b, *z, tt[6];

  F<double,80> Fv, *FL;
  B<double>    Bv, *BL, *BL2;

  Nmat = 40;
  N = Nmat+40;

  L   = new double[N];
  L_b = new double[N];
  v_d = new double[2*N];
  L_d = new double[2*N*N];
  L2  = new double[N*N];
  L3  = new double[N*N];

  FL  = new F<double,80>[N];
  BL  = new B<double>[N];
  BL2 = new B<double>[N];

  lambda   = new double[N];
  lambda_b = new double[N];
  z        = new double[N];

  //----------------------------------------------------------//
  //                                                          //
  // First, test portfolio sensitivity on its own             //
  //                                                          //
  //----------------------------------------------------------//

  for(i=0;i<N;i++) L[i] = 0.05;
  portfolio_b(N,Nmat,delta,Nopt,maturities,swaprates,L,L_b,v);

  for(i=0;i<N*N;i++) L_d[i] = 0;
  for(i=0;i<N;i++) L_d[i+i*N] = 1;
  portfolio_d(N,Nmat,delta,Nopt,maturities,swaprates,L,L_d,v,v_d);

  for(i=0;i<N;i++) BL[i] = 0.05;
  portfolio(N,Nmat,delta,Nopt,maturities,swaprates,BL,Bv);
  Bv.diff(0,1);

  for(i=0;i<N;i++) {
    FL[i] = 0.05;
    FL[i].diff(i);
    // FL[i].diff(i,N);
  }
  portfolio(N,Nmat,delta,Nopt,maturities,swaprates,FL,Fv);

  printf("\n portfolio sensitivity check \n\n");
  for (i=0; i<N; i++)
    printf(" %f  %f  %f  %f\n",L_b[i],v_d[i],BL[i].d(0),Fv.d(i));

  //----------------------------------------------------------//
  //                                                          //
  // Next, do a full path + portfolio sensitivity             //
  // and a timing study                                       //
  //                                                          //
  // A real application would generate a different random     //
  // vector z for each path but here we set one and reuse it  //
  //                                                          //
  //----------------------------------------------------------//

  for (i=0; i<Nmat; i++) z[i]=0.3;

  for (pass=0; pass<6; pass++) {

    if (pass==0)
      npath = 1;
    else if (pass<4)
      npath = 100000;
    else
      npath = 10000;

    for (path=0; path<npath; path++) {

      if (pass==1) {
        // plain evaluation for timing comparison //
        for(i=0;i<N;i++) {
          L[i] = 0.05;
          lambda[i] = 0.2;
        }
        path_calc(N,Nmat,delta,L,lambda,z);
        portfolio(N,Nmat,delta,Nopt,maturities,swaprates,L,v);
      }
      if (pass==0 | pass==2) {
        // hand-coded reverse mode //
        for(i=0;i<N;i++) {
          L[i] = 0.05;
          lambda[i] = 0.2;
        }
        for(i=0;i<N;i++) L[i] = 0.05;
        path_calc_b1(N,Nmat,delta,L,lambda,z,L2,L3);
        portfolio_b(N,Nmat,delta,Nopt,maturities,swaprates,L,L_b,v);
        path_calc_b2(N,Nmat,delta,L,L_b,lambda,lambda_b,z,L2,L3);
      }

      if (pass==0 | pass==3) {
        // hand-coded forward mode //
        for(i=0;i<N;i++) L[i] = 0.05;
        for(i=0;i<N*N;i++) L_d[i] = 0;
        for(i=0;i<N;i++) L_d[i+i*N] = 1;
        path_calc_d(N,Nmat,delta,L,L_d,lambda,z);
        portfolio_d(N,Nmat,delta,Nopt,maturities,swaprates,L,L_d,v,v_d);
      }

      if (pass==0 | pass==4) {
        // AD reverse mode //
        for(i=0;i<N;i++) {
          BL[i] = 0.05;
          BL2[i] = BL[i];  // copy to avoid changing independent vars
          lambda[i] = 0.2;
        }
        path_calc(N,Nmat,delta,BL2,lambda,z);
        portfolio(N,Nmat,delta,Nopt,maturities,swaprates,BL2,Bv);
        for(i=0;i<N;i++) {
          BL2[i] = 0;      // zero out so no longer dependent
        }
        Bv.diff(0,1);
      }

      if (pass==0 | pass==5) {
        // AD forward mode //
        for(i=0;i<N;i++) {
          FL[i] = 0.05;
          FL[i].diff(i);
          // FL[i].diff(i,N);
          lambda[i] = 0.2;
        }
        path_calc(N,Nmat,delta,FL,lambda,z);
        portfolio(N,Nmat,delta,Nopt,maturities,swaprates,FL,Fv);
      }
    }

    if (pass==0) {
      printf("\n path + portfolio sensitivity check \n\n");
      for (i=0; i<N; i++)
        printf(" %f  %f  %f  %f\n",L_b[i],v_d[i],BL[i].d(0),Fv.d(i));
    }

    tt[pass] = ((double) clock())/CLOCKS_PER_SEC;
    if (pass>0) tt[pass-1] = (tt[pass]-tt[pass-1])*(1000.0/npath);
  }

  printf("\n timings in milliseconds per path\n");
  printf("   val  hc-b  hc-f   BAD   FAD \n");
  printf(" %5.2f %5.2f %5.2f %5.2f %5.2f\n",tt[0],tt[1],tt[2],tt[3],tt[4]);

  return 0;
}
