//======================================================================
//
//                    vtlmdriver.cpp
//
// Purpose:
//          - verify, compare and time TAC++ generated tangent linear
//            code (vector mode) against hand-written code by Mike Giles.
// 
//======================================================================

/* Program to compute deltas and vegas of swaption portfolio
   from forward and reverse mode pathwise sensitivities 
   written by Mike Giles in 2005-7 based on code written by
   Zhao and Glasserman at Columbia University */


#include <stdlib.h>
#include <stdio.h>
#include <time.h>


#include "tac++-codes.h"
#include "mikes-codes.h"


#undef VERIFY_PORTFOLIO_VTLM  // debugging only


/* -------------------------------------------------------- */

int main(){

    double delta = 0.25; /* LIBOR interval  */

    // data for swaption portfolio //
    int    Nopt = 15;
    int    maturities[] = {4,4,4,8,8,8,20,20,20,28,28,28,40,40,40};
    double swaprates[]  = {.045,.05,.055,.045,.05,.055,.045,.05,
			   .055,.045,.05,.055,.045,.05,.055 };

    int       i, j, N, Nmat, pass, path, npath;
    double    *v, *v_d, *L, *L_b, *L_d, *L2, *L3, 
	*lambda, *lambda_b, *z, tt[6];

    //------------------------------
    // NOTE:
    // - the tangent linear code must
    //   have been generated with option
    //   "-jacobian pjacmax" with
    //   pjacmax equal to the N below!
    //------------------------------
    Nmat = 40;
    N = Nmat+40;

    v   = (double*) malloc(sizeof(double));
    L   = (double*) malloc(N*sizeof(double));
    L_b = (double*) malloc(N*sizeof(double));
    v_d = (double*) malloc(2*N*sizeof(double));
    L_d = (double*) malloc(2*N*N*sizeof(double));
    L2  = (double*) malloc(N*N*sizeof(double));
    L3  = (double*) malloc(N*N*sizeof(double));

    lambda   = (double*) malloc(N*sizeof(double));
    lambda_b = (double*) malloc(N*sizeof(double));
    z        = (double*) malloc(N*sizeof(double));


    //------------------------------
    // variables required for running
    // tac(++) modifications
    //
    // MVCOMMENT:
    // - C++ reference type variables
    //   are not yet supported by tac(++).
    //   Thus, we must introduce pointer
    //   type variables for "output" arugments
    //   (2 vars are required for cost2 check)
    //------------------------------
    double *vptr;             // function value
    double *v_vtl;            // tac++ tlm derivatives (output)
    double L_vtl[N][pjacmax]; // tac++ seed matrix     (input to TLM routine)
    vptr  = (double*) malloc(sizeof(double));
    v_vtl = (double*) malloc(N*sizeof(double));

    //----------------------------------------------------------//
    //                                                          //
    // First, test portfolio sensitivity on its own             //
    //                                                          //
    //----------------------------------------------------------//

    for(i=0;i<N;i++) L[i] = 0.05;
    portfolio_b(N,Nmat,delta,Nopt,maturities,swaprates,L,L_b,v);

    for(i=0;i<N*N;i++) L_d[i] = 0;
    for(i=0;i<N;i++) L_d[i+i*N] = 1;
    portfolio_d(N,Nmat,delta,Nopt,maturities,swaprates,L,L_d,v,v_d);

    //============================================================
    //
    //          verify portfolio TLM sensitivites
    //
    // Note:
    //       - debugging only
    //
    //============================================================
#ifdef VERIFY_PORTFOLIO_VTLM
    for(i=0;i<N;i++) L[i] = 0.05;
    // setup diagonal seed matrix
    for(i=0; i<N; i++) {
	for(j=0; j<N; j++) {
	    L_vtl[i][j] = 0;
	    if(i==j)
		L_vtl[i][j] = 1.;
	}
	v_vtl[i] = 0.;
    }

    tac_portfolio_vtl(N, Nmat, delta, Nopt, maturities, swaprates, L, L_vtl, vptr, v_vtl);


    printf("\n==================================================\n");
    printf("  portfolio sensitivity check, hand vs. tac(++)\n");
    printf("==================================================\n");
      
    printf( "%4s %14s  %14s  %14s  \n",
	    "idx", "hand TLM", "TAC(++) TLM ", "rel. Diff");
    for(i=0; i<N; i++) {
	printf("%4d  %14.8f  %14.8f  %14.8f\n", i, v_d[i], v_vtl[i], reldiff(v_d[i],v_vtl[i], epsmach));
    }
#endif  
  
  
    //----------------------------------------------------------//
    //                                                          //
    // Next, do a full path + portfolio sensitivity             //
    // and a timing study                                       //
    //                                                          //
    // A real application would generate a different random     //
    // vector z for each path but here we set one and reuse it  //
    //                                                          //
    //----------------------------------------------------------//

    for (i=0; i<Nmat; i++) z[i]=0.3;


    for (pass=0; pass<6; pass++) {
	if (pass==0) {
	    npath = 1;
	}
	else if (pass<6) {
#ifdef NPATHTIMING
	    npath = NPATHTIMING;
#else
	    npath = 1000;
#endif
	}
	for (path=0; path<npath; path++) {

	    if (pass==1) {
		// plain evaluation for timing comparison //
		for(i=0;i<N;i++) {
		    L[i] = 0.05;
		    lambda[i] = 0.2;
		}
		path_calc(N,Nmat,delta,L,lambda,z);
		portfolio(N,Nmat,delta,Nopt,maturities,swaprates,L,v);
	    }
	    if (pass==0 | pass==2) {
		// hand-coded reverse mode //
		for(i=0;i<N;i++) {
		    L[i] = 0.05;
		    lambda[i] = 0.2;
		}
		for(i=0;i<N;i++) L[i] = 0.05;
		path_calc_b1(N,Nmat,delta,L,lambda,z,L2,L3);
		portfolio_b(N,Nmat,delta,Nopt,maturities,swaprates,L,L_b,v);
		path_calc_b2(N,Nmat,delta,L,L_b,lambda,lambda_b,z,L2,L3);
	    }

	    if (pass==0 | pass==3) {
		// hand-coded forward mode //
		for(i=0;i<N;i++) L[i] = 0.05;
		for(i=0;i<N*N;i++) L_d[i] = 0;
		for(i=0;i<N;i++) L_d[i+i*N] = 1;
		path_calc_d(N,Nmat,delta,L,L_d,lambda,z);
		portfolio_d(N,Nmat,delta,Nopt,maturities,swaprates,L,L_d,v,v_d);
	    }

	    if( pass==4 ) {
		// plain evaluation for timing comparison //
		for(i=0;i<N;i++) {
		    L[i] = 0.05;
		    lambda[i] = 0.2;
		}
		model(N, Nmat, delta, Nopt, maturities, z, swaprates, L, lambda, vptr);
	    }

	    if( pass==0 | pass==5 ) {
		for(i=0; i<N; i++) {
		    L[i] = 0.05;
		    lambda[i] = 0.2;
		    for(j=0; j<N; j++) {
			L_vtl[i][j] = 0;
			if(i==j)
			    L_vtl[i][j] = 1.;
		    }
		    v_vtl[i] = 0.;
		}
		model_vtl(N, Nmat, delta, Nopt, maturities, z, swaprates, L, L_vtl, lambda, vptr, v_vtl);
	    }

	}

	if (pass==0) {

	    printf("\n==================================================\n");
	    printf("  path + portfolio sensitivity check, hand vs. tac(++) (vector mode)\n");
	    printf("==================================================\n");
	    printf( "%4s %14s  %14s  %14s  \n",
		    "idx", "hand TLM", "TAC(++) TLM ", "rel. Diff");
	    for(i=0; i<N; i++) {
		printf("%4d %14.8f  %14.8f  %14.8f\n", i, v_d[i], v_vtl[i], reldiff(v_d[i],v_vtl[i], epsmach));
	    }
	}

	tt[pass] = ((double) clock())/CLOCKS_PER_SEC;
	if (pass>0) tt[pass-1] = (tt[pass]-tt[pass-1])*(1000.0/npath);
    }

    printf("\n==================================================\n");
    printf("          TIMINGS:\n");
    printf("==================================================\n");
    printf("  absolute timings in milliseconds per path (based on %d paths)\n", npath);
    printf(" val       : pathcalc+portfolio evaluation (Mikes original)\n");
    printf(" hc-b      : Mikes hand-coded reverse\n");
    printf(" hc-f      : Mikes hand-coded forward\n");
    printf(" tac++-val : pathcalc+portfolio evalutation (TAC++ compliant version)\n");
    printf(" tac++-vf  : TAC++ generated tangent linear (vector-mode)\n");
    printf("-----------------------------------------------------------------------------\n");
    printf(" %10s  %10s  %10s  %10s  %10s\n", "val", "hc-b", "hc-f", "tac++-val", "tac++-vf");
    printf(" %10.2f %10.2f %10.2f %10.2f %10.2f\n", tt[0], tt[1], tt[2], tt[3], tt[4]);
    printf("-----------------------------------------------------------------------------\n");
    printf("  relative timings\n");
    printf(" adjoint(hand)/func           : %5.2f\n", tt[1]/tt[0]);
    printf(" tangent(hand)/func           : %5.2f\n", tt[2]/tt[0]);
    printf(" TLM(TAC++, vector-mode)/func : %5.2f\n", tt[4]/tt[3]);

    return 0;
}
